import arzumify.polyenergy.impl.CoordinateMatchMaker;
import net.minecraft.util.math.Vec3i;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.Timeout;

import static org.junit.jupiter.api.Assertions.assertTrue;

@Timeout(1)
public class EnergyTests {
    public static void assertWithError(double expected, double actual, double error) {
        assertTrue(expected - error <= actual && actual <= expected + error);
    }

    public static void assertWithError(double expected, double actual) {
        assertWithError(expected, actual, 1);
    }

    @BeforeEach
    public void setup() {
        // Delete all batteries
        CoordinateMatchMaker.reset();
        System.out.println("Reset CoordinateMatchMaker");
    }

    @Test
    public void testEnergyTransfer() {
        System.out.println("Energy transfer test");
        // Create two batteries with coordinates (0, 0, 0) and (1, 0, 0)
        // Battery1 should be able to provide energy to battery2
        var battery1 = new SimpleCodeOnlyBattery(new Vec3i(0, 0, 0), 16, 0, 1, "Battery1");
        var battery2 = new SimpleCodeOnlyBattery(new Vec3i(1, 0, 0), 16, 1, 0, "Battery2");

        // Give battery1 some energy
        battery1.energy = 16;

        // Tick the batteries 10 times
        for (int i = 0; i < 10; i++) {
            SimpleCodeOnlyBattery.tick(battery1);
            SimpleCodeOnlyBattery.tick(battery2);
        }

        // Output the energy of the batteries
        System.out.println("Battery1 energy: " + battery1.energy);
        System.out.println("Battery2 energy: " + battery2.energy);

        // Battery1 should have lost 10 energy and battery2 should have gained 10 energy
        assertWithError(6, battery1.energy);
        assertWithError(10, battery2.energy);
    }

    @Test
    public void testUnloadBattery1() {
        System.out.println("Unloading battery1 test");
        // Create two batteries with coordinates (0, 0, 0) and (1, 0, 0)
        // Battery1 should be able to provide energy to battery2
        var battery1 = new SimpleCodeOnlyBattery(new Vec3i(0, 0, 0), 16, 0, 1, "Battery1");
        var battery2 = new SimpleCodeOnlyBattery(new Vec3i(1, 0, 0), 16, 1, 0, "Battery2");

        // Give battery1 some energy
        battery1.energy = 16;

        // Tick 5 times
        for (int i = 0; i < 5; i++) {
            SimpleCodeOnlyBattery.tick(battery1);
            SimpleCodeOnlyBattery.tick(battery2);
        }

        // Unload battery1
        System.out.println("Unloading battery1");
        battery1.onUnload(null, null);

        // Tick 5 more times
        for (int i = 0; i < 5; i++) {
            SimpleCodeOnlyBattery.tick(battery2);
        }

        // Output the energy of the batteries
        System.out.println("Battery1 energy: " + battery1.energy);
        System.out.println("Battery2 energy: " + battery2.energy);

        // Battery1 should have lost 6 energy and battery2 should have gained 6 energy (±1)
        assertWithError(10, battery1.energy);
        assertWithError(6, battery2.energy);
    }

    @Test
    public void testUnloadBattery2() {
        System.out.println("Unloading battery2 test");
        // Create two batteries with coordinates (0, 0, 0) and (1, 0, 0)
        // Battery1 should be able to provide energy to battery2
        var battery1 = new SimpleCodeOnlyBattery(new Vec3i(0, 0, 0), 16, 0, 1, "Battery1");
        var battery2 = new SimpleCodeOnlyBattery(new Vec3i(1, 0, 0), 16, 1, 0, "Battery2");

        // Give battery1 some energy
        battery1.energy = 16;

        // Tick 5 times
        for (int i = 0; i < 5; i++) {
            SimpleCodeOnlyBattery.tick(battery1);
            SimpleCodeOnlyBattery.tick(battery2);
        }

        // Unload battery2
        System.out.println("Unloading battery2");
        battery2.onUnload(null, null);

        // Tick 5 more times
        for (int i = 0; i < 5; i++) {
            SimpleCodeOnlyBattery.tick(battery1);
            SimpleCodeOnlyBattery.tick(battery2);
        }

        // Output the energy of the batteries
        System.out.println("Battery1 energy: " + battery1.energy);
        System.out.println("Battery2 energy: " + battery2.energy);

        // Battery1 should have lost 5 energy and battery2 should have gained 5 energy
        assertWithError(11, battery1.energy);
        assertWithError(5, battery2.energy);
    }

    @Test
    public void testUnloadThenReloadBattery1() {
        System.out.println("Unloading then reloading battery1 test");
        // Create two batteries with coordinates (0, 0, 0) and (1, 0, 0)
        // Battery1 should be able to provide energy to battery2
        var battery1 = new SimpleCodeOnlyBattery(new Vec3i(0, 0, 0), 16, 0, 1, "Battery1");
        var battery2 = new SimpleCodeOnlyBattery(new Vec3i(1, 0, 0), 16, 1, 0, "Battery2");

        // Give battery1 some energy
        battery1.energy = 16;

        // Tick 5 times
        for (int i = 0; i < 5; i++) {
            SimpleCodeOnlyBattery.tick(battery1);
            SimpleCodeOnlyBattery.tick(battery2);
        }

        // Unload battery1
        System.out.println("Unloading battery1");
        battery1.onUnload(null, null);

        // Tick 5 more times
        for (int i = 0; i < 5; i++) {
            SimpleCodeOnlyBattery.tick(battery2);
        }

        // Reload battery1
        System.out.println("Reloading battery1");
        battery1.onLoad(null, null);

        // Tick 5 more times
        for (int i = 0; i < 5; i++) {
            SimpleCodeOnlyBattery.tick(battery1);
            SimpleCodeOnlyBattery.tick(battery2);
        }

        // Output the energy of the batteries
        System.out.println("Battery1 energy: " + battery1.energy);
        System.out.println("Battery2 energy: " + battery2.energy);

        // Battery1 should have lost 10 energy and battery2 should have gained 10 energy
        assertWithError(6, battery1.energy);
        assertWithError(10, battery2.energy);
    }

    @Test
    public void testUnloadThenReloadBattery2() {
        System.out.println("Unloading then reloading battery2 test");
        // Create two batteries with coordinates (0, 0, 0) and (1, 0, 0)
        // Battery1 should be able to provide energy to battery2
        var battery1 = new SimpleCodeOnlyBattery(new Vec3i(0, 0, 0), 16, 0, 1, "Battery1");
        var battery2 = new SimpleCodeOnlyBattery(new Vec3i(1, 0, 0), 16, 1, 0, "Battery2");

        // Give battery1 some energy
        battery1.energy = 16;

        // Tick 5 times
        for (int i = 0; i < 5; i++) {
            SimpleCodeOnlyBattery.tick(battery1);
            SimpleCodeOnlyBattery.tick(battery2);
        }

        // Unload battery2
        System.out.println("Unloading battery2");
        battery2.onUnload(null, null);

        // Tick 5 more times
        for (int i = 0; i < 5; i++) {
            SimpleCodeOnlyBattery.tick(battery1);
            SimpleCodeOnlyBattery.tick(battery2);
        }

        // Reload battery2
        System.out.println("Reloading battery2");
        battery2.onLoad(null, null);

        // Tick 5 more times
        for (int i = 0; i < 5; i++) {
            SimpleCodeOnlyBattery.tick(battery1);
            SimpleCodeOnlyBattery.tick(battery2);
        }

        // Output the energy of the batteries
        System.out.println("Battery1 energy: " + battery1.energy);
        System.out.println("Battery2 energy: " + battery2.energy);

        // Battery1 should have lost 10 energy and battery2 should have gained 10 energy
        assertWithError(6, battery1.energy);
        assertWithError(10, battery2.energy);
    }

    @Test
    public void twoHalfFullBatteries() {
        System.out.println("Two half-full batteries");
        // Create two batteries with coordinates (0, 0, 0) and (1, 0, 0)
        // They should both provide energy to each other and stay at equilibrium
        var battery1 = new SimpleCodeOnlyBattery(new Vec3i(0, 0, 0), 16, 1, 1, "Battery1");
        var battery2 = new SimpleCodeOnlyBattery(new Vec3i(1, 0, 0), 16, 1, 1, "Battery2");

        // Give battery1 and battery2 some energy
        battery1.energy = 8;
        battery2.energy = 8;

        // Tick 11 times
        for (int i = 0; i < 11; i++) {
            SimpleCodeOnlyBattery.tick(battery1);
            SimpleCodeOnlyBattery.tick(battery2);
        }

        // Output the energy of the batteries
        System.out.println("Battery1 energy: " + battery1.energy);
        System.out.println("Battery2 energy: " + battery2.energy);

        // Battery1 and battery2 should have lost no energy
        assert (battery1.energy >= 7 && battery1.energy <= 9);
        assert (battery2.energy >= 7 && battery2.energy <= 9);
    }

    @Test
    public void middleOneCharges() {
        System.out.println("Middle one charges");
        // Create three batteries with coordinates (0, 0, 0), (1, 0, 0), and (2, 0, 0)
        // Battery2 should provide energy to battery1 and battery3
        var battery1 = new SimpleCodeOnlyBattery(new Vec3i(0, 0, 0), 16, 1, 0, "Battery1");
        var battery2 = new SimpleCodeOnlyBattery(new Vec3i(1, 0, 0), 16, 0, 1, "Battery2");
        var battery3 = new SimpleCodeOnlyBattery(new Vec3i(2, 0, 0), 16, 1, 0, "Battery3");

        // Give battery2 some energy
        battery2.energy = 16;

        // Tick 10 times
        for (int i = 0; i < 10; i++) {
            SimpleCodeOnlyBattery.tick(battery1);
            SimpleCodeOnlyBattery.tick(battery2);
            SimpleCodeOnlyBattery.tick(battery3);
        }

        // Output the energy of the batteries
        System.out.println("Battery1 energy: " + battery1.energy);
        System.out.println("Battery2 energy: " + battery2.energy);
        System.out.println("Battery3 energy: " + battery3.energy);

        // Battery1 and battery3 should have gained 8 energy and battery2 should have lost 16 energy
        assertWithError(8, battery1.energy);
        assertWithError(0, battery2.energy);
        assertWithError(8, battery3.energy);
    }

    @Test
    public void middleOneReceives() {
        System.out.println("Middle one receives");
        // Create three batteries with coordinates (0, 0, 0), (1, 0, 0), and (2, 0, 0)
        // Battery1 and battery3 should provide energy to battery2
        var battery1 = new SimpleCodeOnlyBattery(new Vec3i(0, 0, 0), 16, 0, 1, "Battery1");
        var battery2 = new SimpleCodeOnlyBattery(new Vec3i(1, 0, 0), 16, 1, 0, "Battery2");
        var battery3 = new SimpleCodeOnlyBattery(new Vec3i(2, 0, 0), 16, 0, 1, "Battery3");

        // Give battery1 and battery3 some energy
        battery1.energy = 8;
        battery3.energy = 8;

        // Tick 17 times
        for (int i = 0; i < 17; i++) {
            SimpleCodeOnlyBattery.tick(battery1);
            SimpleCodeOnlyBattery.tick(battery2);
            SimpleCodeOnlyBattery.tick(battery3);
        }

        // Output the energy of the batteries
        System.out.println("Battery1 energy: " + battery1.energy);
        System.out.println("Battery2 energy: " + battery2.energy);
        System.out.println("Battery3 energy: " + battery3.energy);

        // Battery2 should have 16 energy and battery1 and battery3 should have 0 energy
        assertWithError(0, battery1.energy);
        assertWithError(16, battery2.energy);
        assertWithError(0, battery3.energy);
    }

    @Test
    public void testEnergyTransferWithRange() {
        System.out.println("Energy transfer with range test");
        // Create two batteries with coordinates (0, 0, 0) and (2, 0, 0)
        // Battery1 should be able to provide energy to battery2
        var battery1 = new SimpleCodeOnlyBattery(new Vec3i(0, 0, 0), 16, 0, 1, "Battery1");
        var battery2 = new SimpleCodeOnlyBattery(new Vec3i(2, 0, 0), 16, 1, 0, "Battery2");

        // Give battery1 some energy
        battery1.energy = 16;

        // Tick the batteries 10 times
        for (int i = 0; i < 10; i++) {
            SimpleCodeOnlyBattery.tick(battery1);
            SimpleCodeOnlyBattery.tick(battery2);
        }

        // Output the energy of the batteries
        System.out.println("Battery1 energy: " + battery1.energy);
        System.out.println("Battery2 energy: " + battery2.energy);

        // Battery1 should have lost no energy and battery2 should have gained no energy
        assertWithError(16, battery1.energy);
        assertWithError(0, battery2.energy);
    }

    @Test
    public void testCables() {
        System.out.println("Cables test");
        // Create two batteries with coordinates (0, 0, 0) and (2, 0, 0) and a cable at (1, 0, 0)
        // Battery1 should be able to provide energy to battery2
        var battery1 = new SimpleCodeOnlyBattery(new Vec3i(0, 0, 0), 16, 0, 1, "Battery1");
        var battery2 = new SimpleCodeOnlyBattery(new Vec3i(2, 0, 0), 16, 1, 0, "Battery2");
        var cable = new SimpleCodeOnlyCable(new Vec3i(1, 0, 0), 1, 1, 1, "Cable");

        // Give battery1 some energy
        battery1.energy = 16;

        // Tick the batteries and cable 16 times
        for (int i = 0; i < 16; i++) {
            SimpleCodeOnlyBattery.tick(battery1);
            SimpleCodeOnlyCable.tick(cable);
            SimpleCodeOnlyBattery.tick(battery2);
        }

        // Output the energy of the batteries
        System.out.println("Battery1 energy: " + battery1.energy);
        System.out.println("Battery2 energy: " + battery2.energy);

        // Output the energy of the cable
        System.out.println("Cable energy: " + cable.energy);

        // Battery1 should have lost 16 energy, battery2 should have gained 15 energy, and the cable should have 1 energy
        assertWithError(1, cable.energy);
        assertWithError(0, battery1.energy);
        assertWithError(15, battery2.energy);
    }

    @Test
    public void longCableChain() {
        System.out.println("Long cable chain test");
        // Create 2 batteries with coordinates (0, 0, 0) and (1001, 0, 0) and 1000 cables in between
        // Battery1 should be able to provide energy to battery2
        var battery1 = new SimpleCodeOnlyBattery(new Vec3i(0, 0, 0), 32, 0, 1, "Battery1");
        var battery2 = new SimpleCodeOnlyBattery(new Vec3i(11, 0, 0), 16, 1, 0, "Battery2");
        var cables = new SimpleCodeOnlyCable[10];
        for (int i = 0; i < 10; i++) {
            cables[i] = new SimpleCodeOnlyCable(new Vec3i(i + 1, 0, 0), 16, 1, 1, "Cable" + (i + 1));
        }

        // Give battery1 some energy
        battery1.energy = 32;

        // Tick the batteries and cables
        for (int i = 0; i < 10; i++) {
            SimpleCodeOnlyBattery.tick(battery1);
            SimpleCodeOnlyBattery.tick(battery2);
            for (SimpleCodeOnlyCable cable : cables) {
                SimpleCodeOnlyCable.tick(cable);
            }
        }

        // Output the energy of the batteries
        System.out.println("Battery1 energy: " + battery1.energy);
        System.out.println("Battery2 energy: " + battery2.energy);

        // Output the energy of the cables
        for (SimpleCodeOnlyCable cable : cables) {
            System.out.println(cable.name + " energy: " + cable.energy);
        }
    }

    @Test
    public void stressTest() {
        System.out.println("Stress test");
        // Create 1000 batteries with coordinates (0, 0, 0) to (999, 0, 0)
        // Each battery should be able to provide energy to the next battery
        var batteries = new SimpleCodeOnlyBattery[1000];
        for (int i = 0; i < 1000; i++) {
            batteries[i] = new SimpleCodeOnlyBattery(new Vec3i(i, 0, 0), 16, 1, 2, "Battery" + i);
        }

        // Give the first battery some energy
        batteries[0].energy = 16;

        // Tick once
        for (SimpleCodeOnlyBattery battery : batteries) {
            SimpleCodeOnlyBattery.tick(battery);
        }
    }
}
